# MR.mpl
# Macaulay Resultant
# (C) 2003-2007 Manfred Minimair
# Licensed under the GPL
# 
# Macaulay Resultant package
# Version 1.4.3 for Maple 9/9.5/10
# 6/30/2008
#
# Please, report bugs or suggestions to manfred@minimair.org.
#
# Basic algorithm: cf. "Using Algebraic Geometry" by Cox, Little, O'Shea
# Only computes the characteristic polynomial when needed.
# Canny, J., "Generalised characteristic polynomials", p.241--250,
#           J. Symbolic Comput. Vol. 9, No. 3, 1990
#
# Work is supported by NSF grant CCF 0430741
########################################################################

# Usage:
# MR:-MResultant(plist, vlist, options)

  # plist # list of inhomogeneous polynomials in the variables in vlist
  # vlist # list of variables, nops(plist) = nops(vlist) + 1
  # Returns the Macaulay resultant of the polynomials in plist.

########################################################################

MR := module()

export MResultant, MacaulayMatrix, Minor, NumeratorMatrix, DenominatorMatrix;

local MRLocal;

option package,
       `Copyright (C) 2007 Manfred Minimair. Version 1.4.1. Licensed under the GNU GENERAL PUBLIC LICENSE, gpl.txt (http://www.gnu.org).`;

########################################################################

MRLocal := module()
export MResultantLocal, NumeratorMatrix, DenominatorMatrix;

local 
IsReduced, MonomialDegree, Parameters,
       MDeterminant, MCharacteristicPolynomial,
       DetQuotient, CharPolyQuotient, MResultantMain,
       MCharPolyQuotient, PolyQuo, ProdExcept;

option package,
       `Copyright (C) 2007 Manfred Minimair. Version 1.4.1. Licensed under the GNU GENERAL PUBLIC LICENSE, gpl.txt (http://www.gnu.org).`;

########################################################################





IsReduced := proc(
  alpha,  # exponent list
  degrees # exponent list of same length as alpha
  )
  # Returns true iff x^alpha is reduced w.r.t. x^degrees, i.e.
  # x^degrees[i] divides x^alpha for exactly one i.

  local i, numDividing;

  option `Copyright (C) 2004 Manfred Minimair.`;

  numDividing := 0;
  i := 1;
  while i <= nops(alpha) and numDividing <= 2 do
    if alpha[i] >= degrees[i] then
      numDividing := numDividing + 1;
    end if;
    i := i + 1;
  end do;

  evalb(numDividing = 1);

end proc;


MonomialDegree := proc(
  degrees # list of total degrees
  )
# Returns the total degree of the monomials used to form the matrices

  local i;

  option `Copyright (C) 2004 Manfred Minimair.`;

  1 + add(degrees[i]-1,i=1..nops(degrees))

end proc;


NumeratorMatrix := proc(
  plist,                # list of polynomials in variables in vlist
  vlist,                # list of variables with nops(plist) = nops(vlist) + 1
  degrees,              # list of degrees of the input polynomials
# OUT:
  exponents::evaln      # list of exponent lists for the monomials 
                        # indexing the returned matrix
                        # the monomials that are contained in the polynomials generated by the procedure
  )::Matrix;
  # Returns the numerator matrix for the Macaulay resultant
# Acknowledgement:
# Function "NumeratorMatrix" based on some code by Petr Lisonek.

  local f, n, x, i, m, j, k, A, row, L, c, fdegrees, dExponents,
        d, exponentsSet, getS, S, rowPoly, rowMult, colIndex;

  option `Copyright (C) 2004 Manfred Minimair.`;

  # Initialize   
  f := plist:
  n := nops(f);
  # xn is the "virtual" homogenization variable
  x := array(1..n);
  for i to n-1 do x[i]:=vlist[i] end do;
  fdegrees := degrees;

  d := MonomialDegree(fdegrees);

  # Compute the list "exponents" of exponent vectors for the monomials of degree d

  dExponents := [ op( combinat[composition](d+n, n) ) ];
  dExponents := map( t -> [seq(t[i]-1,i=1..n)] , dExponents );

# Partition the monomials of degree d into disjoint lists S[i]:
# The disjoint lists are concatenated as the list "dExponents".
# In order to be able to recover S[i] from "dExponents", 
# an array getS is computed, 
# where i = getS[k] if exponets[k] is in S[i].
  exponentsSet := {op(dExponents)};
  S := array(1..n);
  for i to n do
    S[i] := select( t -> (t[i]>=fdegrees[i]) , exponentsSet);
    exponentsSet := exponentsSet minus S[i];
  end do;
  dExponents := [ seq( op(S[i]),i=1..n) ];
  exponents := dExponents;

  m := nops(dExponents); # size of the square numerator matrix
  getS := array(1..m);
  j := 0:
  for i from 1 to n do
    for k from j+1 to j+nops(S[i]) do
      getS[k] := i:
    end do:
    j := j+nops(S[i]):
  end do:

  # Compute numerator matrix

  A := array(1..m,1..m);
  for row from 1 to m do

    i := getS[row]:

    # Compute monomial exponent vector for the multiplier monomial for each row
    L := [seq(0,j=1..n)]: 
    L[i] := fdegrees[i]:
    rowMult := dExponents[row]-L:

    # Compute coefficients of row
    rowPoly := expand( mul( x[j]^rowMult[j], j=1..n-1) * f[i] );
    for k from 1 to m do
      colIndex := dExponents[k];

      # Compute coefficient of x^colIndex
      c := rowPoly;
      for j from 1 to n-1 do
        c := coeff(c, x[j], colIndex[j])
      end do;
      
      A[row,k] := c 
    end do;

  end do;

  A := convert(A, Matrix);
  A;
end proc;

DenominatorMatrix := proc(
  numeratorMatrix::Matrix, # Macaulay resultant numerator matrix
  exponents,               # list of exponent lists indexing the numerator matrix
  degrees,                 # total degrees of the polynomials used to generate numeratorMatrix
# OUT:
  unreduced::evaln         # list of "unreduced" exponent lists
  )::Matrix;
  # Returns the denominator matrix for the Macaulay resultant

  local i, m, unreducedExponents, denominatorMatrix;

  option `Copyright (C) 2004 Manfred Minimair.`;

  m := LinearAlgebra[RowDimension](numeratorMatrix);

  # Compute denominator matrix
  unreducedExponents := [];
  for i from 1 to m do
    if not(IsReduced(exponents[i],degrees)) then
      unreducedExponents := [op(unreducedExponents), i];
    end if;
  end do;
  unreduced := unreducedExponents;

  #printf("Size of the denominator matrix: %d\n", nops(unreducedExponents));

  denominatorMatrix := LinearAlgebra[SubMatrix](numeratorMatrix, unreducedExponents, unreducedExponents);

  denominatorMatrix;
end proc;

MDeterminant := proc(
  DetFun, # proc(M::Matrix) used to compute determinants
  M::Matrix
  ) 
  # Returns the determinant of M. If M=[] then its determinant is 1.

  local zDet;

  option `Copyright (C) 2004 Manfred Minimair.`;

  if LinearAlgebra[RowDimension](M) = 0 then
    zDet := 1
  else
    zDet := DetFun(M)
  end if;
  zDet;
end proc;

MCharacteristicPolynomial := proc(
  CharPolyFun, # proc(M::Matrix,v) used to compute the 
               # characteristic polynomial
  M::Matrix,
  v    # variable for the characteristic polynomial of M
  ) 
  # Returns the characteristic polynomial of M. 
  # If M=[] then its characteristic polynomial is 1.

  local zCharPoly;

  option `Copyright (C) 2004 Manfred Minimair.`;

  if LinearAlgebra[RowDimension](M) = 0 then
    zCharPoly := 1
  else
    #print("M", M);
    #print("v", v);
    zCharPoly := CharPolyFun(M, v);
  end if;
  zCharPoly;
end proc;

DetQuotient := proc(
  mNumer,    # numerator
  mDenom,    # denominator
  indetList # list of indeterminates in mNumer, mDenom
  )
  # Returns the quotient mNumer/mDenom

  local mQuo, r;

  option `Copyright (C) 2004 Manfred Minimair.`;

  if indetList = [] then
    if type(mNumer, integer) and type(mDenom, integer) then
      mQuo := iquo(mNumer, mDenom, 'r')
    else
      mQuo := mNumer/mDenom;
      r := 1
    end if;
  else
    mQuo := quo(mNumer, mDenom, op(1,indetList), 'r')
  end if;
  mQuo
end proc;

MCharPolyQuotient := proc(
  QuoFun,         # proc(numer, denom, v) used to compute the
                  # quotient,remainder of polynomials in the variable v
  mNumer,         # numerator characteristic polynomial
  mDenom,         # denominator characteristic polynomial
  lambda::uneval  # variable of the characteristic polynomial
  )
  # Returns the quotient mNumer/mDenom

  local mQuoPoly, r, mQuo;

  option `Copyright (C) 2004 Manfred Minimair.`;
  mQuoPoly,r := QuoFun(mNumer, mDenom, lambda);
  mQuo := coeff(mQuoPoly, lambda, 0);
  mQuo
end proc;

PolyQuo := proc(
  numer, # numerator in variable v
  denom, # denominator in variable v
  v
  )
  # Returns a pair: exact quotient of numer/denom, remainder

  local r, q;

  option `Copyright (C) 2004 Manfred Minimair.`;

  q := quo(numer, denom, v, 'r');
  return q, r;
end proc;

MResultantMain := proc(
  plist, # list of inhomogeneous polynomials in the variables in vlist
  vlist, # list of variables, nops(plist) = nops(vlist) + 1
  params, # list of parameters of the input polynomials
  lambda, # variable for characteristic polynomial
  degrees # list of total degrees of polynomials in plist
  )
  # Returns the Macaulay resultant of the polynomials in plist.

  local numerMatrix, denomMatrix, fdegrees, exponents, unreduced,
        mNumer, mDenom, mRemainder, r, mQuo, macaulayResultant,
        DeterminantFun, CharacteristicPolynomialFun, QuotientFun;

  option `Copyright (C) 2005 Manfred Minimair.`;

#print("plist", plist);
#print("vlist", vlist);
#print("params", params);
#print("lambda", lambda);
#print("degrees", degrees);

  DeterminantFun := LinearAlgebra:-Determinant;
  CharacteristicPolynomialFun := LinearAlgebra:-CharacteristicPolynomial;
  QuotientFun := PolyQuo;
  fdegrees := degrees;

  numerMatrix := NumeratorMatrix(plist,vlist, fdegrees, exponents);
  denomMatrix := DenominatorMatrix(numerMatrix, exponents, fdegrees, unreduced);
  mDenom := MDeterminant(DeterminantFun, denomMatrix);

#print("mDenom", mDenom);

  if mDenom <> 0 then 
    mNumer := MDeterminant(DeterminantFun, numerMatrix);
    macaulayResultant := DetQuotient(mNumer, mDenom, params);
  else
#print("numerMatrix", numerMatrix);
    mNumer := MCharacteristicPolynomial(CharacteristicPolynomialFun, numerMatrix, lambda);
    mDenom := MCharacteristicPolynomial(CharacteristicPolynomialFun, denomMatrix, lambda);
    macaulayResultant := MCharPolyQuotient(QuotientFun,mNumer, mDenom, lambda);
  end if;

  macaulayResultant;
end proc;

Parameters := proc(
  plist, # list of inhomogeneous polynomials in the variables in vlist
  vlist # list of variables
  )
  # Returns the list of parameters in plist, 
  # i.e. indeterminates minus vlist

  local vset, i, params;

  option `Copyright (C) 2004 Manfred Minimair.`;


  vset := {op(vlist)};
  params := {};
  for i from 1 to nops(plist) do
    params := params union (indets(op(i,plist)) minus vset);
  end do;

  [op(params)];

end proc;


MResultantLocal := proc(
  plist, # list of inhomogeneous polynomials in the variables in vlist
  vlist # list of variables, nops(plist) = nops(vlist) + 1
  )
  # Returns the Macaulay resultant of the polynomials in plist.

  local flist, params, lambda, i, computedDegrees;

  option `Copyright (C) 2005 Manfred Minimair.`;

  flist := plist;
  if nops(flist) <> nops(vlist)+1 then
    printf("Number of polynomials and variables do not match!\n");
    return FAIL;
  end if;

  params := Parameters(flist, vlist);

  computedDegrees := [];
  for i from 1 to nops(flist) do
     computedDegrees := [op(computedDegrees), degree(flist[i], {op(vlist)})];
  end do;

  MResultantMain(args[1], args[2], params, lambda, computedDegrees);

  end proc;

end module;
########################################################################

# Export MResultantLocal:
##########################

MResultant := proc()
  option `Copyright (C) 2004 Manfred Minimair.`;
  MRLocal:-MResultantLocal(args);
end proc;

NumeratorMatrix := proc()
  option `Copyright (C) 2004 Manfred Minimair.`;
  MRLocal:-NumeratorMatrix(args);
end proc;

MacaulayMatrix := proc()
  option `Copyright (C) 2004 Manfred Minimair.`;
  MRLocal:-NumeratorMatrix(args);
end proc;

DenominatorMatrix := proc()
  option `Copyright (C) 2004 Manfred Minimair.`;
  MRLocal:-DenominatorMatrix(args);
end proc;

Minor := proc()
  option `Copyright (C) 2004 Manfred Minimair.`;
  MRLocal:-DenominatorMatrix(args);
end proc;

end module;
########################################################################
